"use strict";

var cmds = require("../cmds");

var _require = require("../index"),
    promisify = _require.promisify;

test("call should return a call cmd", function () {
  var fn = /*#__PURE__*/regeneratorRuntime.mark(function fn() {
    return regeneratorRuntime.wrap(function fn$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
          case "end":
            return _context.stop();
        }
      }
    }, fn, this);
  });
  var actual = cmds.call(fn, "foo", "bar");
  var expected = {
    type: "call",
    fn: fn,
    args: ["foo", "bar"]
  };
  expect(actual).toEqual(expected);
});

test("call should run validator", function () {
  var fn = /*#__PURE__*/regeneratorRuntime.mark(function fn() {
    return regeneratorRuntime.wrap(function fn$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
          case "end":
            return _context2.stop();
        }
      }
    }, fn, this);
  });
  fn.validator = function () {
    throw new Error("oops");
  };
  try {
    cmds.call(fn, "foo", "bar");
  } catch (e) {
    expect(e.message).toEqual("oops");
    return;
  }
  fail("call did not throw");
});

test("call should run validator on promisified function", function () {
  var fn = /*#__PURE__*/regeneratorRuntime.mark(function fn() {
    return regeneratorRuntime.wrap(function fn$(_context3) {
      while (1) {
        switch (_context3.prev = _context3.next) {
          case 0:
          case "end":
            return _context3.stop();
        }
      }
    }, fn, this);
  });
  fn.validator = function () {
    throw new Error("oops");
  };
  try {
    cmds.call(promisify(fn), "foo", "bar");
  } catch (e) {
    expect(e.message).toEqual("oops");
    return;
  }
  fail("call did not throw");
});

test("callFn should return a callFn cmd", function () {
  var fn = function fn() {};
  var actual = cmds.callFn(fn, "foo", "bar");
  var expected = {
    type: "callFn",
    fn: fn,
    args: ["foo", "bar"]
  };
  expect(actual).toEqual(expected);
});

test("call.fn === callFn", function () {
  expect(cmds.call.fn).toEqual(cmds.callFn);
});

test("callFnBound should return a callFnBound cmd", function () {
  var fn = function fn() {};
  var bindThis = {};
  var actual = cmds.callFnBound(bindThis, fn, "foo", "bar");
  var expected = {
    type: "callFn",
    fn: fn,
    bindThis: bindThis,
    args: ["foo", "bar"]
  };
  expect(actual).toEqual(expected);
});

test("call.fnBound === callFnBound", function () {
  expect(cmds.call.fnBound).toEqual(cmds.callFnBound);
});

test("callCallback should return a callCallback cmd", function () {
  var fn = function fn() {};
  var bindThis = {};
  var actual = cmds.callCallback(fn, "foo", "bar");
  var expected = {
    type: "callCallback",
    fn: fn,
    args: ["foo", "bar"]
  };
  expect(actual).toEqual(expected);
});

test("call.callback === callCallback", function () {
  expect(cmds.call.callback).toEqual(cmds.callCallback);
});

test("callCallbackBound should return a callCallbackBound cmd", function () {
  var fn = function fn() {};
  var bindThis = {};
  var actual = cmds.callCallbackBound(bindThis, fn, "foo", "bar");
  var expected = {
    type: "callCallback",
    fn: fn,
    bindThis: bindThis,
    args: ["foo", "bar"]
  };
  expect(actual).toEqual(expected);
});

test("call.callbackBound === callCallbackBound", function () {
  expect(cmds.call.callbackBound).toEqual(cmds.callCallbackBound);
});

test("echo()", function () {
  expect(cmds.echo("foo")).toEqual({ type: "echo", message: "foo" });
});

test("globalVariable()", function () {
  expect(cmds.globalVariable("process")).toEqual({
    type: "globalVariable",
    name: "process"
  });
});

test("globalVariable() should throw key is falsey", function () {
  expect(function () {
    return cmds.globalVariable();
  }).toThrow("name required");
});

test("now()", function () {
  expect(cmds.now()).toEqual({ type: "now" });
});

test("log()", function () {
  expect(cmds.log("foo", "bar")).toEqual({ type: "log", args: ["foo", "bar"] });
});

test("logError()", function () {
  expect(cmds.logError("foo", "bar")).toEqual({
    type: "logError",
    args: ["foo", "bar"]
  });
});

test("setImmediate()", function () {
  var cmd = cmds.now();
  expect(cmds.setImmediate(cmd)).toEqual({ type: "setImmediate", cmd: cmd });
});

test("setImmediate() should clear stack, if set", function () {
  var cmd = cmds.now();
  expect(cmds.setImmediate(cmd, true)).toEqual({
    type: "setImmediate",
    cmd: cmd,
    resetStack: true
  });
});

test("setTimeout()", function () {
  var cmd = cmds.now();
  expect(cmds.setTimeout(cmd, 1000)).toEqual({
    type: "setTimeout",
    cmd: cmd,
    time: 1000
  });
});

test("setTimeout() should clear stack, if set", function () {
  var cmd = cmds.now();
  expect(cmds.setTimeout(cmd, 1000, true)).toEqual({
    type: "setTimeout",
    cmd: cmd,
    time: 1000,
    resetStack: true
  });
});

test("clearTimeout()", function () {
  var id = 123;
  expect(cmds.clearTimeout(id)).toEqual({
    type: "clearTimeout",
    id: id
  });
});

test("setInterval()", function () {
  var cmd = cmds.now();
  expect(cmds.setInterval(cmd, 1000)).toEqual({
    type: "setInterval",
    cmd: cmd,
    time: 1000
  });
});

test("setInterval() should clear stack, if set", function () {
  var cmd = cmds.now();
  expect(cmds.setInterval(cmd, 1000, true)).toEqual({
    type: "setInterval",
    cmd: cmd,
    time: 1000,
    resetStack: true
  });
});

test("clearInterval()", function () {
  var id = 123;
  expect(cmds.clearInterval(id)).toEqual({
    type: "clearInterval",
    id: id
  });
});

test("sleep()", function () {
  expect(cmds.sleep(1000)).toEqual({ type: "sleep", time: 1000 });
});

test("series()", function () {
  var cmdList = [];
  expect(cmds.series(cmdList)).toEqual({ type: "series", cmdList: cmdList });
});

test("parallel()", function () {
  var cmdList = [];
  expect(cmds.parallel(cmdList)).toEqual({ type: "parallel", cmdList: cmdList });
});

test("envelope()", function () {
  var cmd = cmds.now();
  expect(cmds.envelope(cmd)).toEqual({ type: "envelope", cmd: cmd });
});

test("either()", function () {
  var cmd = cmds.now();
  var defaultValue = 0;
  expect(cmds.either(cmd, defaultValue)).toEqual({
    type: "either",
    cmd: cmd,
    defaultValue: defaultValue
  });
});

test("getState()", function () {
  expect(cmds.getState("foo")).toEqual({
    type: "getState",
    path: "foo"
  });
});

test("getState() with default value", function () {
  expect(cmds.getState("foo", "bar")).toEqual({
    type: "getState",
    path: "foo",
    defaultValue: "bar"
  });
});

test("setState()", function () {
  expect(cmds.setState("foo", "bar")).toEqual({
    type: "setState",
    path: "foo",
    value: "bar"
  });
});

test("clearState()", function () {
  expect(cmds.clearState()).toEqual({
    type: "clearState"
  });
});
//# sourceMappingURL=cmds.spec.js.map