"use strict";

var assert = require("assert");

var _require = require("../interpreters"),
    setImmediate = _require.setImmediate,
    setTimeout = _require.setTimeout,
    setInterval = _require.setInterval,
    globalVariable = _require.globalVariable;

var _require2 = require("../core"),
    call = _require2.call;

var _require3 = require("./common"),
    interpreters = _require3.interpreters,
    functions = _require3.functions,
    cmds = _require3.cmds;

var _require4 = require("./test-util"),
    sleep = _require4.sleep;

var coreInterpreters = require("../interpreters");
var coreCmds = require("../cmds");

test("setImmediate() should swallow errors to prevent node uncaught rejection errors", async function () {
  var context = {};
  // Should not cause a node unhandled rejection error
  await setImmediate({ cmd: cmds.die() }, { call: call, context: context, interpreters: interpreters });
  // wait for setImmediate to finish to get code coverage
  await sleep(10);
});

test("setImmediate() should reset stack, if resetStack is truthy", async function () {
  var _marked = /*#__PURE__*/regeneratorRuntime.mark(nestedCall);

  var count = 0;
  var stack = void 0;
  var onCommand = function onCommand(t) {
    if (t.command.type !== "call") return;
    count++;
    if (count >= 10) {
      stack = t.context.stack;
    }
  };
  function nestedCall() {
    return regeneratorRuntime.wrap(function nestedCall$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            if (!(count >= 10)) {
              _context.next = 2;
              break;
            }

            return _context.abrupt("return");

          case 2:
            _context.next = 4;
            return coreCmds.setImmediate(coreCmds.call(nestedCall), true);

          case 4:
          case "end":
            return _context.stop();
        }
      }
    }, _marked, this);
  }
  var context = { onCommand: onCommand, name: "telemetry" };
  var now = Date.now();
  await call(context, coreInterpreters, nestedCall, "foo");
  await sleep(50);
  expect(stack).toBeTruthy();
  expect(stack.length).toEqual(1);
});

test("setTimeout() should swallow errors to prevent node uncaught rejection errors", async function () {
  var context = {};
  // Should not cause a node unhandled rejection error
  await setTimeout({ cmd: cmds.die(), time: 0 }, { call: call, context: context, interpreters: interpreters });
  // wait for setTimeout to finish to get code coverage
  await sleep(10);
});

test("setTimeout() should reset stack, if resetStack is truthy", async function () {
  var _marked2 = /*#__PURE__*/regeneratorRuntime.mark(nestedCall);

  var count = 0;
  var stack = void 0;
  var onCommand = function onCommand(t) {
    if (t.command.type !== "call") return;
    count++;
    if (count >= 10) {
      stack = t.context.stack;
    }
  };
  function nestedCall() {
    return regeneratorRuntime.wrap(function nestedCall$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            if (!(count >= 10)) {
              _context2.next = 2;
              break;
            }

            return _context2.abrupt("return");

          case 2:
            _context2.next = 4;
            return coreCmds.setTimeout(coreCmds.call(nestedCall), 0, true);

          case 4:
          case "end":
            return _context2.stop();
        }
      }
    }, _marked2, this);
  }
  var context = { onCommand: onCommand, name: "telemetry" };
  var now = Date.now();
  await call(context, coreInterpreters, nestedCall, "foo");
  await sleep(50);
  expect(stack).toBeTruthy();
  expect(stack.length).toEqual(1);
});

test("setInterval() should swallow errors to prevent node uncaught rejection errors", async function () {
  var context = {};
  // Should not cause a node unhandled rejection error
  var id = await setInterval({ cmd: cmds.die(), time: 0 }, { call: call, context: context, interpreters: interpreters });
  // wait for setInterval to finish to get code coverage
  await sleep(10);
  clearInterval(id);
});

test("setInterval() should reset stack, if resetStack is truthy", async function () {
  var _marked3 = /*#__PURE__*/regeneratorRuntime.mark(nestedCall);

  var count = 0;
  var stack = void 0;
  var onCommand = function onCommand(t) {
    if (t.command.type !== "now") return;
    count++;
    if (count >= 10) {
      stack = t.context.stack;
    }
  };
  function nestedCall() {
    return regeneratorRuntime.wrap(function nestedCall$(_context3) {
      while (1) {
        switch (_context3.prev = _context3.next) {
          case 0:
            if (!(count >= 10)) {
              _context3.next = 2;
              break;
            }

            return _context3.abrupt("return");

          case 2:
            _context3.next = 4;
            return coreCmds.setInterval(coreCmds.now(), 0, true);

          case 4:
          case "end":
            return _context3.stop();
        }
      }
    }, _marked3, this);
  }
  var context = { onCommand: onCommand, name: "telemetry" };
  var now = Date.now();
  await call(context, coreInterpreters, nestedCall, "foo");
  await sleep(50);
  expect(stack).toBeTruthy();
  expect(stack.length).toEqual(1);
});

test("globalVariable() should use window if available", function () {
  window.foo = "bar";
  var value = globalVariable({ name: "foo" });
  assert.equal(value, "bar");
});
//# sourceMappingURL=interpreters.spec.js.map