"use strict";

var _require = require("../core"),
    call = _require.call;

var _require2 = require("./common"),
    interpreters = _require2.interpreters,
    cmds = _require2.cmds,
    functions = _require2.functions;

var basic = functions.basic,
    basicMultistep = functions.basicMultistep,
    basicParallel = functions.basicParallel,
    basicMultistepParallel = functions.basicMultistepParallel,
    basicEmpty = functions.basicEmpty,
    basicMultiArg = functions.basicMultiArg;


test("should be able to process a command", async function () {
  var expected = "foo";
  var actual = await call({}, interpreters, basic, expected);
  expect(actual).toEqual(expected);
});

test("should be able to accept array arguments", async function () {
  var expected = ["foo", "bar"];
  var actual = await call({}, interpreters, basic, expected);
  expect(actual).toEqual(expected);
});

test("should be able to handle multiple arguments", async function () {
  var actual = await call({}, interpreters, basicMultiArg, "foo", "bar");
  expect(actual).toEqual("foobar");
});

test("should be able to process commands in multiple steps", async function () {
  var actual = await call({}, interpreters, basicMultistep, "foo");
  var expected = { s1: "foo1", s2: "foo2" };
  expect(actual).toEqual(expected);
});

test("should be able to process commands in parallel", async function () {
  var actual = await call({}, interpreters, basicParallel, "foo");
  var expected = { s1: "foo1", s2: "foo2" };
  expect(actual).toEqual(expected);
});

test("should be able to process commands in parallel, in multiple steps", async function () {
  var actual = await call({}, interpreters, basicMultistepParallel, "foo");
  var expected = { s1: "foo1", s2: "foo2", s3: "foo3", s4: "foo4" };
  expect(actual).toEqual(expected);
});

test("should be able to process an empty list of commands", async function () {
  var expected = [];
  var actual = await call({}, interpreters, basicEmpty);
  expect(actual).toEqual(expected);
});

test("interpreters should be able to return non-promise values", async function () {
  var _marked = /*#__PURE__*/regeneratorRuntime.mark(valueReturningInterpreter);

  function valueReturningInterpreter(value) {
    return regeneratorRuntime.wrap(function valueReturningInterpreter$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            _context.next = 2;
            return cmds.echo(value);

          case 2:
            return _context.abrupt("return", _context.sent);

          case 3:
          case "end":
            return _context.stop();
        }
      }
    }, _marked, this);
  }

  var actual = await call({}, interpreters, valueReturningInterpreter, "bar");
  var expected = "bar";
  expect(actual).toEqual(expected);
});

test("interpreters should be able to return promises", async function () {
  var _marked2 = /*#__PURE__*/regeneratorRuntime.mark(promiseReturningInterpreter);

  function promiseReturningInterpreter(value) {
    return regeneratorRuntime.wrap(function promiseReturningInterpreter$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            _context2.next = 2;
            return cmds.echoPromise(value);

          case 2:
            return _context2.abrupt("return", _context2.sent);

          case 3:
          case "end":
            return _context2.stop();
        }
      }
    }, _marked2, this);
  }

  var actual = await call({}, interpreters, promiseReturningInterpreter, "bar");
  var expected = "bar";
  expect(actual).toEqual(expected);
});

test("interpreters should be able to return an array of results", async function () {
  var fn = /*#__PURE__*/regeneratorRuntime.mark(function fn(a, b) {
    var result;
    return regeneratorRuntime.wrap(function fn$(_context3) {
      while (1) {
        switch (_context3.prev = _context3.next) {
          case 0:
            _context3.next = 2;
            return [cmds.echo(a), cmds.echo(b)];

          case 2:
            result = _context3.sent;
            return _context3.abrupt("return", result);

          case 4:
          case "end":
            return _context3.stop();
        }
      }
    }, fn, this);
  });
  var actual = await call({}, interpreters, fn, ["foo", "bar"], ["foo", "baz"]);
  var expected = [["foo", "bar"], ["foo", "baz"]];
  expect(actual).toEqual(expected);
});
//# sourceMappingURL=scenarios.spec.js.map